#include "graphDist.h"

void getSymP(bool* adjMat, int num, Eigen::MatrixXf *P){
    int i, j, k, count;

    Eigen::MatrixXf D(num, num);
    Eigen::MatrixXf A(num, num);

    for(i = 0; i < num; ++i){
        for(j = 0; j < num; ++j){

            A(i, j) = 0.0f;

            if(adjMat[i * num + j] && i != j){
                A(i, j) = 1.0f;
            }

            D(i, j) = 0.0f;

            if(i == j){
                count = 0;

                for(k = 0; k < num; ++k){
                    // two options, the number of vertices as degrees
                    //  the summation of weights as degrees
                    if(adjMat[i * num + k] && i != k){
                        count++;
                    }
                }

                D(i, j) = 1.0f / count;
            }
        }
    }

    Eigen::MatrixXf halfD(num, num);
    halfD = D.sqrt();

    Eigen::MatrixXf temp(num, num);
    temp = halfD * A;

    (*P) = temp * halfD;
}

void convertA2String(bool* adjMat, int num, vector<int>* myString, Eigen::MatrixXf *P){
    int i, j, k;

    getSymP(adjMat, num, P);

    // Find the largest eigen value of P.
    Eigen::EigenSolver<Eigen::MatrixXf> es((*P));

    float real, largest;
    int largestId;

    largest = -FLT_MAX;
    complex<float> lambda;

    for(i = 0; i < num; ++i){
        lambda = es.eigenvalues()[i];
        real = lambda.real();

        if(largest < real){
            largest = real;
            largestId = i;
        }
    }

    // Fetch its corresponding eigenvector E
    // we know that the size of the eigenvector is num
    vector<pair<int, float> > elements;

    for(i = 0; i < num; ++i){
        pair<int, float> p;
        p.first = i;
        p.second = es.eigenvectors().col(largestId).coeff(i).real();

        elements.push_back(p);
    }

    sortDecreasOrder mySortDecreasOrder;
    sort(elements.begin(), elements.end(), mySortDecreasOrder);

    bool* visited = new bool[num];

    memset(visited, false, sizeof(bool) * num);

    int largestElementId;
    int tempLargestElementId;

    for(i = 0; i < num; ++i){
        // we know there are num of nodes
        // thus there are num iterations

        if(i == 0){
            largestElementId = elements[i].first;

            (*myString).push_back(largestElementId);

            visited[largestElementId] = true;
        }

        // we will go through the neighbors of largestElementId
        // find the next largest
        tempLargestElementId = -1;

        // elements are sorted
        for(j = 1; j < num; ++j){
            if(adjMat[largestElementId * num + elements[j].first] &&
                    visited[elements[j].first] == false){

                    tempLargestElementId = j;
            }
        }

        if(tempLargestElementId == -1){
            // there is no nearest neighbor,
            // then we need to find the next largest one
            // who is the neighbor of existing id
            bool find = false;

            for(j = 1; j < num && find == false; ++j){
                if(visited[elements[j].first] == false){
                    for(k = 0; k < num && find == false; ++k){
                        if(visited[k] && adjMat[k * num + elements[j].first]){
                            tempLargestElementId = j;

                            find = true;
                        }
                    }
                }
            }
        }

        if(tempLargestElementId != -1){
            largestElementId = elements[tempLargestElementId].first;
            visited[largestElementId] = true;
            (*myString).push_back(largestElementId);
        }else{
            break;
        }
    }

    delete[] visited;
}

void getDegrees(bool* adjMat, vector<unsigned int> *vec_degree, unsigned int num_nodes){
    unsigned int i, j, d;

    for(i = 0; i < num_nodes; ++i){
        d = 0;

        for(j = 0; j < num_nodes; ++j){
            if(i != j && adjMat[i * num_nodes + j]){
                d += 1;
            }
        }

        (*vec_degree).push_back(d);
    }
}

double beta(int di, int dj){
    int max, min;

    if(di > dj){
        max = di;
        min = dj;
    }else{
        max = dj;
        min = di;
    }

    return 1.0 * (max - min) / max;
}

double nul(int s1, int s2){
    return 2.0 * abs(s1 - s2) / (s1 + s2);
}

double dist_ij(unsigned int num_nodes1, unsigned int num_nodes2,
               vector<unsigned int> vec_degree1, vector<unsigned int> vec_degree2,
               bool* adjMat1, bool* adjMat2,
               Eigen::MatrixXf P1, Eigen::MatrixXf P2,
                      pair<int, int> p1, pair<int, int> p2){


    int a, b, c, d;
    a = p1.first;
    b = p1.second;
    c = p2.first;
    d = p2.second;

    double beta_ab = beta(vec_degree1[a], vec_degree2[b]);

    double beta_cd = beta(vec_degree1[c], vec_degree2[d]);

    double Rd;

    if(adjMat1[a * num_nodes1 + c]){
        Rd = P1(a, c);
    }else{
        Rd = nul(num_nodes1, num_nodes2);
    }

    double Rm;

    if(adjMat2[b * num_nodes2 + d]){
        Rm = P2(b, d);
    }else{
        Rm = nul(num_nodes1, num_nodes2);
    }

    float v = beta_ab * beta_cd * Rd * Rm;

    return v;
}

pair<int, int> findNext(float** dijDis, bool** visited, int num1, int num2){
    int l = -1;
    int r = -1;

    float min = FLT_MAX;

    int i, j;
    for(i = 0; i < num1; ++i){
        for(j = 0; j < num2; ++j){
            if(visited[i][j] == false){
                if(min > dijDis[i][j]){
                    l = i;
                    r = j;

                    min = dijDis[i][j];
                }
            }
        }
    }

    pair<int, int> p;
    p.first = l;
    p.second = r;

    return p;
}

float calDist(unsigned int num_nodes1, unsigned int num_nodes2,
                 vector<unsigned int> vec_degree1, vector<unsigned int> vec_degree2,
                 bool* adjMat1, bool* adjMat2,
                 Eigen::MatrixXf P1, Eigen::MatrixXf P2){
    float** dijDis_temp;
    bool** visited_temp;

    dijDis_temp = new float*[num_nodes1];
    visited_temp = new bool*[num_nodes1];

    int i, j;
    for(i = 0; i < num_nodes1; ++i){
        dijDis_temp[i] = new float[num_nodes2];
        visited_temp[i] = new bool[num_nodes2];
    }

    for(i = 0; i < num_nodes1; ++i){
        for(j = 0; j < num_nodes2; ++j){
            dijDis_temp[i][j] = FLT_MAX;
            visited_temp[i][j] = false;
        }
    }

    dijDis_temp[0][0] = 0.0f;

    //////////////////////////////
    pair<int, int> currNode;
    currNode.first = 0;
    currNode.second = 0;

    int unvisitedSize = num_nodes1 * num_nodes2;

    float tempDis, tempValue;

    while(unvisitedSize > 0){

        // right
        if(currNode.first < num_nodes1 - 1){
            pair<int, int> neighbor;
            neighbor.first = currNode.first + 1;
            neighbor.second = currNode.second;

            tempValue = dist_ij(num_nodes1, num_nodes2, vec_degree1, vec_degree2,
                                adjMat1, adjMat2, P1, P2, currNode, neighbor);

            tempDis = dijDis_temp[currNode.first][currNode.second] +
                    tempValue;

            if(tempDis < dijDis_temp[neighbor.first][neighbor.second]){
                dijDis_temp[neighbor.first][neighbor.second] = tempDis;
            }
        }

        // down
        if(currNode.second < num_nodes2 - 1){
            pair<int, int> neighbor;
            neighbor.first = currNode.first;
            neighbor.second = currNode.second + 1;

            tempValue = dist_ij(num_nodes1, num_nodes2, vec_degree1, vec_degree2,
                                adjMat1, adjMat2, P1, P2, currNode, neighbor);

            tempDis = dijDis_temp[currNode.first][currNode.second] +
                    tempValue;

            if(tempDis < dijDis_temp[neighbor.first][neighbor.second]){
                dijDis_temp[neighbor.first][neighbor.second] = tempDis;
            }
        }

        // right down
        if(currNode.first < num_nodes1 - 1 && currNode.second < num_nodes2 - 1){
            pair<int, int> neighbor;
            neighbor.first = currNode.first + 1;
            neighbor.second = currNode.second + 1;

            tempValue = dist_ij(num_nodes1, num_nodes2, vec_degree1, vec_degree2,
                                adjMat1, adjMat2, P1, P2, currNode, neighbor);

            tempDis = dijDis_temp[currNode.first][currNode.second] +
                    tempValue;

            if(tempDis < dijDis_temp[neighbor.first][neighbor.second]){
                dijDis_temp[neighbor.first][neighbor.second] = tempDis;
            }
        }

        visited_temp[currNode.first][currNode.second] = true;

        if(currNode.first == num_nodes1 - 1 && currNode.second == num_nodes2 - 1){
            break;
        }

        unvisitedSize--;
        currNode = findNext(dijDis_temp, visited_temp, num_nodes1, num_nodes2);
    }

    //////////////////////////////
    for(i = 0; i < num_nodes1; ++i){
        delete[] dijDis_temp[i];
    }

    delete[] dijDis_temp;

    for(i = 0; i < num_nodes1; ++i){
        delete[] visited_temp[i];
    }

    delete[] visited_temp;

    //////////////////////////////
    return dijDis_temp[num_nodes1 - 1][num_nodes2 - 1];
}

float calCommDist(bool* adjMat1, unsigned int num_nodes1,
             bool* adjMat2, unsigned int num_nodes2){

    Eigen::MatrixXf P1(num_nodes1, num_nodes1);
    Eigen::MatrixXf P2(num_nodes2, num_nodes2);

    vector<int> s1;
    vector<int> s2;

    convertA2String(adjMat1, num_nodes1, &s1, &P1);
    convertA2String(adjMat2, num_nodes2, &s2, &P2);

    vector<unsigned int> vec_degree1;
    vector<unsigned int> vec_degree2;

    getDegrees(adjMat1, &vec_degree1, num_nodes1);
    getDegrees(adjMat2, &vec_degree2, num_nodes2);

    float d = calDist(num_nodes1,  num_nodes2,
                vec_degree1, vec_degree2,
                adjMat1, adjMat2, P1, P2);

    return d;
}

void readGraph(char* fn, bool* adjMat, unsigned int num_nodes2){
    FILE* pFile = fopen(fn, "rb");

    if(pFile != NULL){
        fread(adjMat, sizeof(bool), num_nodes2, pFile);

        fclose(pFile);
    }else{
        printf("Cannot open file %s\n", fn);
    }
}

void storeDist(char* fn, float v){
    FILE* pFile = fopen(fn, "wt");

    if(pFile != NULL){
        fprintf(pFile, "%f \n", v);

        fclose(pFile);
    }else{
        printf("Cannot open file %s\n", fn);
    }
}

int main(int argc, char* argv[]){
    char ifn1[1024];
    char ifn2[1024];
    char ofn[1024];

    unsigned int num_nodes1, num_nodes2, num_nodes12, num_nodes22;

    if (argc != 6) {
        // Tell the user how to run the program
        printf("Input Error. Please input: the input and output data path, number of nodes, number of iterations and C2.");

        return 1;
    }else{
        for (unsigned int i = 1; i < 6; ++i) {
            // Remember argv[0] is the path to the program, we want from argv[1] onwards
           if(i == 1){
               strcpy(ifn1, argv[i]);
           }else if(i == 2){
               strcpy(ifn2, argv[i]);
           }else if(i == 3){
               strcpy(ofn, argv[i]);
           }else if(i == 4){
               num_nodes1 = atoi(argv[i]);
           }else if(i == 5){
               num_nodes2 = atoi(argv[i]);
           }
       }
    }

    num_nodes12 = num_nodes1 * num_nodes1;
    num_nodes22 = num_nodes2 * num_nodes2;

    bool* adjMat1 = new bool[num_nodes12];
    bool* adjMat2 = new bool[num_nodes22];

    readGraph(ifn1, adjMat1, num_nodes12);
    readGraph(ifn2, adjMat2, num_nodes22);

    float dist = calCommDist(adjMat1, num_nodes1, adjMat2, num_nodes2);

    storeDist(ofn, dist);

    return 1;
}
