/////////////////////
/// author Yi Gu
///
/// Graph Similarity Calculation:
///
/// Given two graphs, we first convert them
/// into two strings. Then the distance between
/// the two strings becomes the distance between
/// two graphs

#include <Eigen/Dense>
#include <Eigen/Sparse>
#include <Eigen/Eigenvalues>
#include <unsupported/Eigen/MatrixFunctions>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <set>
#include <queue>
#include <map>
#include <vector>
#include <algorithm>

using namespace std;

//------------------------------------------------------------
/// This structure is used to sort a vector in decreasing
/// order
struct sortDecreasOrder {
  bool operator() (pair<int, float> i, pair<int, float> j){
      return (i.second > j.second);
  }
};

//------------------------------------------------------------
/// This function calculates the normalized symmetric random
/// walks probability matrix P
/// adjMat: adjacent matrix
/// num: number of nodes
/// P: the probability matrix
void getSymP(bool* adjMat, int num, Eigen::MatrixXf *P);

//------------------------------------------------------------
/// This function converts a graph to a string
/// adjMat: adjacent matrix
/// num: number of nodes
/// myString: the order of the graph nodes, which forms the required
/// string
/// P: the probability matrix
void convertA2String(bool* adjMat, int num, vector<int>* myString, Eigen::MatrixXf *P);

//------------------------------------------------------------
/// This function calculates the degree of each node
/// adjMat: adjacent matrix
/// vec_degree: the degrees of all the nodes
/// num_nodes: number of nodes
void getDegrees(bool* adjMat, vector<unsigned int> *vec_degree, unsigned int num_nodes);

//------------------------------------------------------------
/// This function calculates the distance between two nodes if
/// they match
/// di and dj: the degrees of the two nodes
double beta(int di, int dj);

//------------------------------------------------------------
/// This function calculates the distance between two nodes if
/// they do not match
/// s1 and s2: the node numbers
double nul(int s1, int s2);

//------------------------------------------------------------
/// This function calculates the distance between two nodes
/// of the two graphs
/// num_nodes1 and num_nodes2: the number of nodes
/// vec_degree1 and vec_degree2: the node degrees
/// adjMat1 and adjMat2: the adjacent matrices
/// P1 and P2: the two probability matrices
/// p1 and p2: the current grid and the next grid locations
double dist_ij(unsigned int num_nodes1, unsigned int num_nodes2,
               vector<unsigned int> vec_degree1, vector<unsigned int> vec_degree2,
               bool* adjMat1, bool* adjMat2,
               Eigen::MatrixXf P1, Eigen::MatrixXf P2,
                      pair<int, int> p1, pair<int, int> p2);

//------------------------------------------------------------
/// This function is a step of the dijkstra's algorithm
/// It looks for the next grid to move
/// dijDist: the distance of dijkstra's algorithm
/// visited: whether the nodes are visited
/// num1 and num2: the numbers of nodes
pair<int, int> findNext(float** dijDis, bool** visited, int num1, int num2);

//------------------------------------------------------------
/// This function calculates the distance between two graphs
/// num_nodes1 and num_nodes2: the number of nodes
/// vec_degree1 and vec_degree2: the node degrees
/// adjMat1 and adjMat2: the adjacent matrices
/// P1 and P2: the two probability matrices
/// p1 and p2: the current grid and the next grid locations
float calDist(unsigned int num_nodes1, unsigned int num_nodes2,
                 vector<unsigned int> vec_degree1, vector<unsigned int> vec_degree2,
                 bool* adjMat1, bool* adjMat2,
                 Eigen::MatrixXf P1, Eigen::MatrixXf P2);

//------------------------------------------------------------
/// This function calculates the necessary resources and calls
/// calDist
/// num_nodes1 and num_nodes2: the number of nodes
/// adjMat1 and adjMat2: the adjacent matrices
float calCommDist(bool* adjMat1, unsigned int num_nodes1,
             bool* adjMat2, unsigned int num_nodes2);

//------------------------------------------------------------
/// This function read an input graph in the form of adjacent matrix
/// fn: file name
/// adjMat: adjacent matrix
/// num_nodes2: num_nodes * num_nodes
void readGraph(char* fn, bool* adjMat, unsigned int num_nodes2);

//------------------------------------------------------------
/// This function stores the distance between the two graphs
void storeDist(char* fn, float v);
