/////////////////////
/// author Yi Gu
///
/// State Clustering Algorithm:
///
/// Starting at a target block,
/// We check its spatial and temporal neighbors,
/// If they are similar to the target, we group
/// them together. Then we continue to check the
/// neighbor's neighbors. This process stops until
/// there is no block similar to the target. Then
/// we find the next block that is not grouped yet,
/// use it as a new target, and repeat the whole
/// process.
///
/// Please note that, if a block is already grouped
/// to a target, we will not check its neighbors in
/// later iterations.
///
///
/// Parallel version: Use a Master-Slave mode
/// Each process contains a set of blocks
///
/// Master first detects which block is the target
/// and broadcast the block ID to all processes
/// All processes load histograms and get their blocks' histograms.
/// Meanwhile, they get the histogram of the target block
///
/// All processes calculate JSD of their own blocks and send
/// the results back to Master. Master use a FIFO queue to
/// traverse the neighborhood and find the next target.
/// If the next target is not in the same time step as
/// the previous one, save the group info to disk

#include <mpi.h>
#include <vector>
#include <stdlib.h>
#include <stddef.h>
#include <math.h>
#include <deque>
#include <assert.h>
#include <string.h>

#include "definition.h"

using namespace std;

//------------------------------------------------------------
/// This function loads the histogram of each block at each time step
/// bins: a storage of all the histograms at one time step
/// t: the index of the time step
void loadData(float* bins, int t);

//------------------------------------------------------------
/// Based on the time step and the index in that time step,
/// this function calculates the corresponding global id of the
/// block
/// t: the index of the time step
/// index: the index of a given block
int getGlobalId(int t, int index);

//------------------------------------------------------------
/// Use JSD as the similarity between two histograms
/// 0.0 means identical, 1.0 mean totally different
/// his1 and his2 are two histograms
float JSD(float his1[BINSIZE], float his2[BINSIZE]);

//------------------------------------------------------------
/// Decide  whether two histograms are similar.
/// Use JSD to calculate the similarity and compare with threshold
/// his1 and his2 are two histograms
/// threshold determines whether two histograms are similar
bool calculateHistogram(float his1[BINSIZE], float his2[BINSIZE], float threshold);

//------------------------------------------------------------
/// go through the queue, and compare each element with the target
/// to decide whether it is similar or not
/// targetHis: the histogram of the target
/// blockQueue: the queue that consists of the blocks
/// targetTime: the time step that the target locates in
/// targetIndex: the index of the target
void compare(float* targetHis, deque<blockInfo> *blockQueue, int targetTime, int targetIndex);

void compare(float* targetHis, deque<blockInfo> *blockQueue);

//------------------------------------------------------------
/// Given a target with its global id
/// We need to calculate which time step it locates in and
/// its index in that time step. In addition, we also want
/// to know the last time step in that time window
/// targetGid: the global id of the target
/// targetTime: the time step that the target locates in
/// targetIndex: the index of the target
/// lastTimeInWindow: the time window
void getTargetTimeAndLastTimeInWindow(int targetGid, int* targetTime,
                                      int* targetIndex, int* lastTimeInWindow);

//------------------------------------------------------------
/// Given a target's global id, remove all the elements
/// with global id less than that global id
/// Use a FIFO mode
/// blockQueue: the queue that consists of the blocks
/// targetGid: the global id of the target
void removeOldBlocks(deque<blockInfo> *blockQueue, int targetGid);

//------------------------------------------------------------
/// If the blockQueue's last element has a time step less than the the last one
/// in the time window, which means, we need more new elements to fill in
/// blockQueue: the queue that consists of the blocks
/// lastTimeInWindow: the time window
/// blockIDs: an array of the block ids
/// num_blocks: the number of the blocks
/// lastLoadedTime: the time steps of the latest uploaded data
void addNewBlocks(deque<blockInfo> *blockQueue, int lastTimeInWindow,
                  int* blockIDs, int num_blocks, int* lastLoadedTime);

//------------------------------------------------------------
/// Check whether this process consists of the target block
/// blockQueue: the queue that consists of the blocks
/// targetTime: the time step that the target locates in
/// targetIndex: the index of the target
/// targetHis: the histogram of the target
bool haveTarget(deque<blockInfo> blockQueue, int targetTime, int targetIndex, float* targetHis);

//------------------------------------------------------------
/// Boradcast so that others know the histogram of the target
/// rank: the index of the process
/// num_proc: the number of processes
/// haveTarget: whether the process has the target
/// targetHis: the histogram of the target
void broadCastHis(int rank, int num_proc, bool haveTarget, float* targetHis);

//------------------------------------------------------------
/// We must report to Master that which elements are similar to our target
/// rank: the index of the process
/// num_proc: the number of processes
/// blockQueue: the queue that consists of the blocks
/// targetTime: the time step that the target locates in
/// isSimilar: array that consists of whether the blocks are similar to the target
void report2Master(int rank, int num_procs, deque<blockInfo> blockQueue, int targetTime, int* isSimilar);

//------------------------------------------------------------
/// Check whether two blocks should be grouped together
/// isSimilar: array that consists of whether the blocks are similar to the target
/// groupIds: array that consists of block ids
bool shouldInAGroup(int* isSimilar, int* groupIds, int offset2, int index_similar);

//------------------------------------------------------------
/// Now Master has collected enough info about the similarity,
/// then we need to group them with region growing algorithm
/// isSimilar: an array to store whether they are similar to the target
/// groupIds: the group ids
/// currGroupIndex: the id of the new group
/// targetGid: the global id of the target
/// targetTime: the time step of the target
/// targetIndex: the index of the target
/// lastTimeInWindow: time window size
void regionGrowing(int* isSimilar, int* groupIds, int roundStarter,
                   int currGroupIndex, int targetGid, int targetTime,
                   int targetIndex, int lastTimeInWindow);

//------------------------------------------------------------
/// Find the next ungrouped block as a new target
/// groupIds: the ids of the groups
/// targetTime: the time step of the target
/// lastTimeInWindow: time window size
/// targetIndex: the index of the target
int findTheNextTarget(int* groupIds, int startRound, int targetTime,
                      int lastTimeInWindow, int targetIndex);

//------------------------------------------------------------
/// save all the targets that we encountered
/// targetRecorder: a vector that stores the target ids
void saveTargets(vector<int> targetRecorder);

//------------------------------------------------------------
/// save the group id of each time step
/// groupIds: the group id of the time step
/// startRound: a parameter
/// targetTime: the time step of the target
/// newT: the time step of the blocks
int saveGroupId(int* groupIds, int startRound, int targetTime, int newT);
